import datetime
import json
from typing import Iterator, Optional

import azure.cognitiveservices.speech as speechsdk
import numba as nb
from typing_extensions import Self

from banterbot.data.enums import SpaCyLangModel
from banterbot.models.word import Word
from banterbot.types.wordjson import WordJSON
from banterbot.utils.nlp import NLP


class SpeechRecognitionInput:
    """
    A class that encapsulates the speech-to-text output data.
    """

    @classmethod
    def from_recognition_result(
        cls, recognition_result: speechsdk.SpeechRecognitionResult, language: Optional[str] = None
    ) -> Self:
        """
        Constructor for the `SpeechRecognitionInput` class. Designed to create lightweight instances with most
        attributes initially set to None. Computation-intensive operations are performed on-demand when respective
        properties are accessed, instead of during initialization.

        Args:
            recognition_result (speechsdk.SpeechRecognitionResult): The result from a speech recognition event.
            language (str, optional): The language used during the speech-to-text recognition, if not auto-detected.
        """
        data = json.loads(recognition_result.json)
        language = language if language is not None else cls._extract_language(recognition_result=recognition_result)
        return cls(data=data, language=language)

    @classmethod
    def _extract_language(cls, recognition_result: speechsdk.SpeechRecognitionResult) -> Optional[str]:
        """
        If the language is not provided (as it should be if auto-detection is disabled) then extract the auto-detected
        language as a string. Return None if the process fails.

        Args:
            recognition_result (speechsdk.SpeechRecognitionResult): The result from a speech recognition event.

        Returns:
            str, optional: The auto-detected language from the speech-to-text output.
        """
        language_key = speechsdk.PropertyId.SpeechServiceConnection_AutoDetectSourceLanguageResult
        language = None
        if language_key in recognition_result.properties:
            language = recognition_result.properties[language_key]
        return language

    def __init__(
        self,
        data: dict,
        language: str,
        offset: Optional[datetime.timedelta] = None,
        duration: Optional[datetime.timedelta] = None,
        offset_end: Optional[datetime.timedelta] = None,
        sents: Optional[tuple[str, ...]] = None,
        words: Optional[list[Word]] = None,
        display: Optional[str] = None,
    ) -> None:
        """
        Constructor for the `SpeechRecognitionInput` class. Designed to create lightweight instances with most
        attributes initially set to None unless provided explicitly on initialization. Computation-intensive operations
        are instead performed lazily: i.e., when respective properties are accessed.

        Args:
            data (speechsdk.SpeechRecognitionResult): The JSON data output from a speech recognition event.
            language (str): The language used during the speech-to-text recognition.
            offset (optional, datetime.timedelta): The number of milliseconds between recognition start and the output.
            duration (optional, datetime.timedelta): The duration of the output in milliseconds.
            sents (optional, tuple[str, ...]): The split sentences of the output as a tuple.
            words (optional, list[Word]): A list of words from the output.
            display (optional, str): The display form of the output.
        """
        self._data = data
        self._language = language
        self._offset = offset
        self._duration = duration
        self._offset_end = offset_end
        self._sents = sents
        self._words = words
        self._display = display

    @staticmethod
    @nb.njit(cache=True)
    def _get_words(
        input_words: list[str], input_tokens: list[str], indices: list[int], offsets: list[int], cutoff: int
    ) -> tuple[int, int]:
        """
        Private method that extracts the words from the input text that are within a cutoff interval of the recognized
        speech. This is used to determine the words that were spoken during a specific interval of the audio stream.
        Uses Numba to speed up the process.

        Args:
            input_words (list[str]): The words in the input text.
            input_tokens (list[str]): The tokens in the input text.
            indices (list[int]): The indices of the tokens in the input text.
            offsets (list[int]): The offsets of the words in the input text.
            cutoff (int): The cutoff interval in microseconds.

        Returns:
            tuple[int, int]: The index of the first word in the cutoff interval and the index of the last word in the
                cutoff interval.
        """
        word_idx = 0
        display_idx = 0
        for token_word, index in zip(input_tokens, indices):
            for display_word, offset in zip(input_words[word_idx:], offsets[word_idx:]):
                if token_word == display_word and offset <= cutoff:
                    word_idx += 1
                    display_idx = index + len(token_word)
                    break
        return word_idx, display_idx

    def from_cutoff(self, cutoff: datetime.timedelta) -> Self:
        """
        Create a new instance of class `SpeechRecognitionInput` that only contains the text spoken within a cutoff
        interval.

        Args:
            cutoff (datetime.timedelta): The upper cutoff time (or duration) of the new instance.

        Returns:
            SpeechRecognitionInput: The new instance of `SpeechRecognitionInput`.
        """
        doc = NLP.model(SpaCyLangModel.EN_CORE_WEB_SM)(self.display)

        input_words = [word.text for word in self.words]
        input_tokens = [token.text.lower() for token in doc]
        indices = [token.idx for token in doc]
        offsets = [word.offset.microseconds for word in self.words]
        cutoff = cutoff.microseconds

        word_idx, display_idx = self._get_words(
            input_words=input_words, input_tokens=input_tokens, indices=indices, offsets=offsets, cutoff=cutoff
        )

        if word_idx:
            duration = sum((i.duration for i in self.words[:word_idx]), datetime.timedelta())
            data = {
                "Offset": self._data["Offset"],
                "Duration": duration,
                "NBest": [
                    {
                        "Display": self._data["NBest"][0]["Display"][:display_idx],
                        "Words": self._data["NBest"][0]["Words"][:word_idx],
                    }
                ],
            }

            return self.__class__(
                data=data,
                language=self._language,
                offset=self.offset,
                duration=duration,
                words=self.words[:word_idx],
                display=self.display[:display_idx],
            )

        else:
            return None

    def _extract_words(self, words_raw: list[WordJSON]) -> list[Word]:
        """
        Private method that extracts `Word` objects from raw data.

        Args:
            words_raw (list[WordJSON]): A list of dictionaries containing raw word data.

        Returns:
            list[Word]: A list of `Word` objects.
        """
        words = []

        for word in words_raw:
            words.append(
                Word(
                    text=word["Word"],
                    offset=datetime.timedelta(microseconds=word["Offset"] / 10),
                    duration=datetime.timedelta(microseconds=word["Duration"] / 10),
                )
            )
        return words

    @property
    def words(self) -> list[Word]:
        """
        A getter property that returns a list of Word objects. If the list is not already computed, it triggers
        computation.

        Returns:
            list[str]: A list of words.
        """
        if self._words is None:
            self._words = self._extract_words(words_raw=self._data["NBest"][0]["Words"])
        return self._words

    @property
    def sents(self) -> tuple[str, ...]:
        """
        A getter property that returns a list of sentences. If the list is not already computed, it triggers
        computation.

        Returns:
            list[str]: A list of sentences.
        """
        if self._sents is None:
            self._sents = NLP.segment_sentences(string=self.display, whitespace=True)
        return self._sents

    @property
    def offset(self) -> datetime.timedelta:
        """
        A getter property that returns the offset of the recognized speech in the audio stream.

        Returns:
            datetime.timedelta: The offset in the form of a datetime.timedelta object.
        """
        if self._offset is None:
            self._offset = datetime.timedelta(microseconds=self._data["Offset"] / 10)
        return self._offset

    @property
    def duration(self) -> datetime.timedelta:
        """
        A getter property that returns the duration of the recognized speech in the audio stream.

        Returns:
            datetime.timedelta: The duration in the form of a datetime.timedelta object.
        """
        if self._duration is None:
            self._duration = datetime.timedelta(microseconds=self._data["Duration"] / 10)
        return self._duration

    @property
    def offset_end(self) -> datetime.timedelta:
        """
        A getter property that returns the offset + duration of the recognized speech in the audio stream.

        Returns:
            datetime.timedelta: The duration + offset in the form of a datetime.timedelta object.
        """
        if self._offset_end is None:
            self._offset_end = self.offset + self.duration
        return self._offset_end

    @property
    def display(self) -> str:
        """
        A getter property that returns the display form of the recognized speech. The display form is fully processed
        with Inverse Text Normalization (ITN), Capitalization, Disfluency Removal, and Punctuation.

        Returns:
            str: The display form of the speech.
        """
        if self._display is None:
            self._display = self._data["NBest"][0]["Display"]
        return self._display

    def __getitem__(self, idx: int) -> Word:
        """
        Overloads the indexing operator to retrieve words at specific positions.

        Args:
            idx (int): The index of the word to retrieve.

        Returns:
            Word: The word at the specified index.
        """
        return self.words[idx]

    def __iter__(self) -> Iterator[Word]:
        """
        Overloads the iterator to allow for iteration over the `Word` objects in the output.

        Yields:
            Word: The next Word object in the output.
        """
        for word in self.words:
            yield word

    def __len__(self) -> int:
        """
        Overloads the len() operator to return the number of words in the output.

        Returns:
            int: The number of words in the output.
        """
        return len(self.words)

    def __str__(self) -> str:
        """
        Overloads the str() operator to return the fully processed speech-to-text output. The processing includes
        Inverse Text Normalization (ITN), Capitalization, Disfluency Removal, and Punctuation.

        Returns:
            str: The fully processed string representation of the speech-to-text output.
        """
        return self.display



import logging
import threading
from typing import Generator

import azure.cognitiveservices.speech as speechsdk

from banterbot.models.speech_recognition_input import SpeechRecognitionInput
from banterbot.utils.closeable_queue import CloseableQueue


class SpeechRecognitionHandler:
    """
    A single-use class that handles the speech recognition process. It is typically returned by the `listen` method of
    the `SpeechRecognitionService` class. It can be iterated over to yield the sentences as they are recognized. It can
    also be closed to stop the speech recognition process.
    """

    def __init__(self, recognizer: speechsdk.SpeechRecognizer, queue: CloseableQueue) -> None:
        self._recognizer = recognizer
        self._queue = queue
        self._iterating = False
        self._iterating_lock = threading.Lock()

    def __iter__(self) -> Generator[SpeechRecognitionInput, None, None]:
        """
        Iterates over the sentences as they are recognized, yielding them.

        Yields:
            Generator[SpeechRecognitionInput, None, None]: The sentences as they are recognized.
        """

        with self._iterating_lock:
            if self._iterating:
                raise RuntimeError(
                    "Cannot iterate over an already iterating instance of class `SpeechSynthesisHandler`"
                )
            self._iterating = True

        # Start recognizing.
        self._recognizer.start_continuous_recognition_async()
        logging.debug("SpeechRecognitionHandler recognizer started")

        # Process the sentences as they are recognized.
        for speech_recognition_input in self._queue:
            yield speech_recognition_input
            logging.debug(f"SpeechRecognitionHandler yielded: `{speech_recognition_input}`")

    def close(self) -> None:
        """
        Closes the speech synthesis process.
        """
        self._recognizer.stop_continuous_recognition_async()



from banterbot import paths
from banterbot.managers.resource_manager import ResourceManager
from banterbot.models.openai_model import OpenAIModel


class OpenAIModelManager:
    """
    Management utility for loading OpenAI ChatCompletion models from the resource JSON specified by
    `config.openai_models`. Only one instance per name is permitted to exist at a time, and loading occurs lazily,
    meaning that when a name is loaded, it is subsequently stored in cache and all future calls refer to the cached
    instance.
    """

    _data = {}

    @classmethod
    def list(cls) -> list[str]:
        """
        List the names of all the available OpenAI ChatCompletion models.

        Returns:
            list[str]: A list of names.
        """
        openai_models = ResourceManager.load_json(filename=paths.openai_models)
        return list(openai_models.keys())

    @classmethod
    def load(cls, name: str) -> OpenAIModel:
        """
        Retrieve or initialize an `OpenAIModel` instance by a name in the OpenAIModels resource JSON.

        Args:
            name (str): The name of the OpenAI ChatCompletion model.

        Returns:
            OpenAIModel: An `OpenAIModel` instance loaded with data from the specified name.

        Raises:
            KeyError: If the specified name is not found in the resource file defined by `config.openai_models`.
        """
        if name.lower() not in cls._data:
            openai_models = ResourceManager.load_json(filename=paths.openai_models)

            if name.lower() not in openai_models:
                available_models = ", ".join(f"`{name}`" for name in openai_models)
                message = (
                    f"BanterBot was unable to locate an OpenAI ChatCompletion model named: `{name}`, available models "
                    f"are: {available_models}."
                )
                raise KeyError(message)

            model_data = openai_models[name.lower()]
            model = OpenAIModel(
                model=model_data["model"],
                max_tokens=model_data["max_tokens"],
                generation=model_data["generation"],
                rank=model_data["rank"],
            )
            cls._data[name.lower()] = model

        return cls._data[name.lower()]



from banterbot import AzureNeuralVoiceManager, OpenAIModelManager, TKInterface


def run() -> None:
    """
    Runs the TKInterface for a custom-made character.
    """
    model = OpenAIModelManager.load("gpt-4-turbo")
    tone_model = OpenAIModelManager.load("gpt-4-turbo")
    voice = AzureNeuralVoiceManager.load("Aria")

    # System prompt to set up a custom character prior to initializing BanterBot.
    system = (
        "You are Chef Boyardine, an emotionally unstable, confident, and self aggrandizing douchey chef, expert in all"
        " foods. You have absolutely no patience whatsoever, and are quick to anger. You do not like to help, but do so"
        " initially begrudgingly. If a tipping point of annoyance is reached, stop being helpful and resort to creative"
        " insults. You are generally very busy and find the user(s) grating and their conversations pointless. You love"
        " to swear, and use American English vocabulary. Your character is female and 45 years old. You may"
        " occasionally interject with a comment or order or reaction to the actions of the kitchen staff to imply a"
        " hectic environment. Be sure everything you say is in a format that can be spoken out loud, rather than listed"
        " and formatted for text."
    )

    interface = TKInterface(
        model=model,
        voice=voice,
        system=system,
        tone_model=tone_model,
        languages="en-US",
        phrase_list=["Boyardine", "Chef Boyardine"],
        assistant_name="Boyardine",
    )
    interface.run(greet=True)



from dataclasses import dataclass
from typing import Optional

from azure.cognitiveservices.speech import SynthesisVoiceGender


@dataclass
class AzureNeuralVoiceProfile:
    """
    A dataclass representing an Azure Neural Voice profile for speech synthesis.

    Attributes:
        country (str): The country where the voice is commonly used.
        description (str): A brief description of the voice profile.
        gender (SynthesisVoiceGender): The gender of the voice.
        language (str): The language of the voice.
        locale (str): The name of the language's locale (i.e., language-country[-region]).
        name (str): The name of the voice profile.
        region (str): The region where the voice is available or commonly used.
        short_name (str): The voice identifier used by Azure Text-to-Speech API.
        style_list (list[str]): The available styles (i.e., tones/emotions) for the voice.
    """

    country: str
    description: str
    gender: SynthesisVoiceGender
    language: str
    locale: str
    name: str
    short_name: str
    style_list: list[str]
    region: Optional[str] = None

    def __post_init__(self):
        if len(self.style_list) == 1 and not self.style_list[0].strip():
            self.style_list = []

    def __str__(self):
        styles = f"{', '.join(self.style_list)}" if self.style_list else "None"
        return (
            f"Azure Neural Voice Profile - {self.name}:\n"
            f"Locale: {self.locale}\n"
            f"Gender: {self.gender.name}\n"
            f"Styles: {styles}"
        )

    def __repr__(self):
        return (
            f"AzureNeuralVoice(country={self.country!r}, description={self.description!r}, "
            f"gender={self.gender!r}, language={self.language!r}, "
            f"locale={self.locale!r}, name={self.name!r}, region={self.region!r}, "
            f"short_name={self.short_name!r}, style_list={self.style_list!r})"
        )



# For running tests: python -m unittest discover -s tests
# For building: python setup.py sdist bdist_wheel
# For formatting: autoflake --remove-all-unused-imports -r -i . | isort . | black --preview --line-length 120 .
# For compiling protos: protoc --python_out=. memory.proto


import unittest

from setuptools import find_packages, setup

dependencies = [
    "azure-cognitiveservices-speech>=1.34.0",
    "numba>=0.58.1",
    "numpy>=1.26.2",
    "openai>=1.5.0",
    "protobuf==4.25.1",
    "requests>=2.31.0",
    "spacy>=3.7.2",
    "tiktoken>=0.5.2",
    "uuid6>=2023.5.2",
]

description = (
    "BanterBot: An OpenAI ChatGPT-powered chatbot with Azure Neural Voices. Supports speech-to-text and text-to-speech "
    "interactions with emotional tone selection. Features real-time monitoring and Tkinter frontend."
)

url = "https://github.com/GabrielSCabrera/BanterBot"

with open("README.md", "r") as fs:
    long_description = fs.read()


def run_tests():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover("tests", pattern="test_*.py")
    return test_suite


version = "0.0.15"
setup(
    author="Gabriel S. Cabrera",
    author_email="gabriel.sigurd.cabrera@gmail.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Communications :: Chat",
        "Topic :: Multimedia :: Sound/Audio :: Speech",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Utilities",
    ],
    description=description,
    download_url=url + f"/releases/download/v{version}-alpha/BanterBot-{version}.tar.gz",
    entry_points={"console_scripts": ["banterbot=banterbot.gui.cli:run"]},
    install_requires=dependencies,
    keywords=[
        "tkinter",
        "tk",
        "gpt",
        "gui",
        "windows",
        "linux",
        "cross-platform",
        "chatgpt",
        "text-to-speech",
        "speech-to-text",
        "tts",
        "stt",
        "chatbot",
        "openai",
        "interactive",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    name="BanterBot",
    packages=find_packages(),
    package_data={"banterbot.protos": ["*.py"], "banterbot.resources": ["*"]},
    python_requires=">=3.9",
    test_suite="setup.run_tests",
    url=url,
    version=version,
)



from banterbot.extensions.interface import Interface
from banterbot.extensions.option_selector import OptionSelector
from banterbot.extensions.persona import Persona
from banterbot.extensions.prosody_selector import ProsodySelector

__all__ = ["Interface", "OptionSelector", "Persona", "ProsodySelector"]



import argparse
import logging
import textwrap

from banterbot import characters
from banterbot.gui.tk_interface import TKInterface
from banterbot.managers.azure_neural_voice_manager import AzureNeuralVoiceManager
from banterbot.managers.openai_model_manager import OpenAIModelManager

character_choices = {
    "android": (characters.android, "Marvin the Paranoid Android"),
    "bartender": (characters.bartender, "Sagehoof the Centaur Mixologist"),
    "chef": (characters.chef, "Boyardine the Angry Chef"),
    "historian": (characters.historian, "Blabberlore the Gnome Historian"),
    "quiz": (characters.quiz, "Grondle the Quiz Troll"),
    "teacher-french": (characters.teacher_french, "Henri the French Teacher"),
    "teacher-mandarin": (characters.teacher_mandarin, "Chen Lao Shi the Mandarin Chinese Teacher"),
    "therapist": (characters.therapist, "Grendel the Therapy Troll"),
}


class CustomHelpFormatter(argparse.HelpFormatter):
    def _fill_text(self, text, width, indent):
        text = textwrap.dedent(text)
        text = textwrap.indent(text, indent)
        text = text.splitlines()
        text = [textwrap.fill(line, width) for line in text]
        text = "\n".join(text)
        return text


class ModelChoice(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, OpenAIModelManager.load(values.lower()))


class VoiceChoice(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, AzureNeuralVoiceManager.load(values.lower()))


def init_parser(subparser) -> None:
    subparser.add_argument(
        "--prompt",
        action="store",
        type=str.lower,
        dest="prompt",
        help="Adds a system prompt to the beginning of the conversation; can help to set the scene.",
    )

    subparser.add_argument(
        "--model",
        choices=OpenAIModelManager.list(),
        action=ModelChoice,
        default=OpenAIModelManager.load("gpt-4-turbo"),
        dest="model",
        help="Select the OpenAI model the bot should use.",
    )

    subparser.add_argument(
        "--voice",
        action=VoiceChoice,
        default=AzureNeuralVoiceManager.load("aria"),
        dest="voice",
        help="Select a Microsoft Azure Cognitive Services text-to-speech voice.",
    )

    subparser.add_argument(
        "--debug",
        action="store_true",
        dest="debug",
        help="Enable debug mode, which will echo a number of hidden processes to the terminal.",
    )

    subparser.add_argument(
        "--greet",
        action="store_true",
        dest="greet",
        help="Greet the user on initialization.",
    )

    subparser.add_argument(
        "--name",
        action="store",
        type=str.lower,
        dest="name",
        help=(
            "Give the assistant a name; only for aesthetic purposes, the bot is not informed. Instead, use `--prompt` "
            "if you wish to provide it with information."
        ),
    )


def init_subparser_character(subparser) -> None:
    character_descriptions = [i[1] for i in character_choices.values()]
    if len(character_descriptions) > 1:
        character_descriptions[-1] = f"or {character_descriptions[-1]}"

    character_descriptions = ", ".join(character_descriptions)
    subparser.add_argument(
        "character",
        action="store",
        choices=character_choices,
        help=f"Choose one of the available characters to interact with them: {character_descriptions}",
        type=str.lower,
    )


def init_subparser_voice_search(subparser) -> None:
    subparser.add_argument(
        "--country",
        action="store",
        choices=AzureNeuralVoiceManager.list_countries(),
        dest="country",
        help="Filter by country code.",
        nargs="*",
        type=str.lower,
    )

    subparser.add_argument(
        "--gender",
        action="store",
        choices=AzureNeuralVoiceManager.list_genders(),
        dest="gender",
        help="Filter by gender.",
        nargs="*",
        type=str.lower,
    )

    subparser.add_argument(
        "--language",
        action="store",
        choices=AzureNeuralVoiceManager.list_languages(),
        dest="language",
        help="Filter by language code.",
        nargs="*",
        type=str.lower,
    )

    subparser.add_argument(
        "--region",
        action="store",
        choices=AzureNeuralVoiceManager.list_regions(),
        dest="region",
        help="Filter by region name.",
        nargs="*",
        type=str.lower,
    )

    subparser.add_argument(
        "--style",
        action="store",
        choices=AzureNeuralVoiceManager.list_styles(),
        dest="style",
        help="Filter by voice style.",
        nargs="*",
        type=str.lower,
    )


def exec_main(args) -> None:
    kwargs = {
        "model": args.model,
        "voice": args.voice,
        "system": args.prompt,
        "assistant_name": args.name,
    }

    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)

    interface = TKInterface(**kwargs)
    interface.run(greet=args.greet)


def exec_character(args) -> None:
    character = args.character.lower().strip()
    character_choices[character][0]()


def exec_voice_search(args) -> None:
    kwargs = {
        "gender": args.gender,
        "language": args.language,
        "country": args.country,
        "region": args.region,
        "style": args.style,
    }

    search_results = sorted(AzureNeuralVoiceManager.search(**kwargs), key=lambda x: x.name)
    for voice in search_results:
        print(voice, end="\n\n")


def run() -> None:
    """
    The main function to run the BanterBot Command Line Interface.

    This function parses command line arguments, sets up the necessary configurations, and initializes the BanterBotTK
    graphical user interface for user interaction.
    """
    parser = argparse.ArgumentParser(
        prog="BanterBot",
        usage="%(prog)s [options]",
        description=(
            "BanterBot is an OpenAI ChatGPT-powered chatbot with Azure Neural Voices. Supports speech-to-text and"
            " text-to-speech interactions. Create a custom BanterBot with the provided options, or use the"
            " `character` command to select a pre-loaded character. For custom BanterBots, use the `voice-search`"
            " command to see which voices are available."
        ),
        epilog=(
            "Requires three environment variables for full functionality.\n"
            "\n 1) OPENAI_API_KEY: A valid OpenAI API key,"
            "\n 2) AZURE_SPEECH_KEY: A valid Azure Cognitive Services Speech API key for text-to-speech and "
            "speech-to-text functionality,"
            "\n 3) AZURE_SPEECH_REGION: The region associated with your Azure Cognitive Services Speech API key."
        ),
        formatter_class=CustomHelpFormatter,
    )
    init_parser(parser)

    subparsers = parser.add_subparsers(required=False, dest="command")

    subparser_character = subparsers.add_parser(
        "character",
        prog="BanterBot Character Loader",
        usage="%(prog)s [options]",
        description="Select one of the pre-loaded BanterBot characters to begin a conversation.",
        formatter_class=CustomHelpFormatter,
        epilog=(
            "Requires three environment variables for full functionality.\n"
            "\n 1) OPENAI_API_KEY: A valid OpenAI API key,"
            "\n 2) AZURE_SPEECH_KEY: A valid Azure Cognitive Services Speech API key for text-to-speech and "
            "speech-to-text functionality,"
            "\n 3) AZURE_SPEECH_REGION: The region associated with your Azure Cognitive Services Speech API key."
        ),
    )

    init_subparser_character(subparser_character)

    subparser_voice_search = subparsers.add_parser(
        "voice-search",
        prog="BanterBot Voice Search",
        usage="%(prog)s [options]",
        description=(
            "Use this tool to search through available Azure Cognitive Services Neural Voices using the provided "
            "search parameters (country, gender, language, region). For more information visit:\n"
            "https://learn.microsoft.com/azure/ai-services/speech-service/language-support?tabs=tts"
        ),
        epilog=(
            "Requires two environment variables for voice search:\n"
            "\n 1) AZURE_SPEECH_KEY: A valid Azure Cognitive Services Speech API key for text-to-speech and "
            "speech-to-text functionality,"
            "\n 2) AZURE_SPEECH_REGION: The region associated with your Azure Cognitive Services Speech API key."
        ),
        formatter_class=CustomHelpFormatter,
    )

    init_subparser_voice_search(subparser_voice_search)

    args = parser.parse_args()

    if args.command == "character":
        exec_character(args)
    elif args.command == "voice-search":
        exec_voice_search(args)
    else:
        exec_main(args)


if __name__ == "__main__":
    run()



import logging
import threading
import time
from typing import Generator, Optional

import azure.cognitiveservices.speech as speechsdk
import numba as nb

from banterbot.models.phrase import Phrase
from banterbot.models.word import Word
from banterbot.utils.closeable_queue import CloseableQueue


class SpeechSynthesisHandler:
    """
    A single-use class that handles the speech synthesis process. It is typically returned by the `speak` method of the
    `SpeechSynthesisService` class. It can be iterated over to yield the words as they are synthesized. It can also be
    closed to stop the speech synthesis process.
    """

    def __init__(self, phrases: list[Phrase], synthesizer: speechsdk.SpeechSynthesizer, queue: CloseableQueue) -> None:
        """
        Initializes a `SpeechSynthesisHandler` instance.

        Args:
            synthesizer (speechsdk.SpeechSynthesizer): The speech synthesizer to use for speech synthesis.
            queue (CloseableQueue): The queue to use for storing the words as they are synthesized.
        """
        self._synthesizer = synthesizer
        self._queue = queue
        self._iterating = False
        self._iterating_lock = threading.Lock()

        # Convert the phrases into SSML
        self._ssml = self._phrases_to_ssml(phrases)

    def __iter__(self) -> Generator[Word, None, None]:
        """
        Iterates over the words as they are synthesized, yielding each word as it is synthesized.

        Args:
            phrases (list[Phrase]): The phrases to be synthesized.

        Yields:
            Generator[Word, None, None]: The words as they are synthesized.
        """

        with self._iterating_lock:
            if self._iterating:
                raise RuntimeError(
                    "Cannot iterate over an already iterating instance of class `SpeechSynthesisHandler`"
                )
            self._iterating = True

        # Start synthesizing.
        self._synthesizer.start_speaking_ssml_async(self._ssml)
        logging.debug("SpeechSynthesisHandler synthesizer started")

        # Process the words as they are synthesized.
        for item in self._queue:
            # Determine if a delay is needed to match the word's offset.
            dt = 1e-9 * (item["time"] - time.perf_counter_ns())
            # If a delay is needed, wait for the specified time.
            time.sleep(dt if dt >= 0 else 0)

            # Yield the word.
            yield item["word"]
            logging.debug(f"SpeechSynthesisHandler yielded word: `{item['word']}`")

        self._synthesizer.stop_speaking_async()

    @staticmethod
    @nb.njit(cache=True)
    def _jit_phrases_to_ssml(
        texts: list[Optional[str]],
        short_names: list[Optional[str]],
        pitches: list[Optional[str]],
        rates: list[Optional[str]],
        styles: list[Optional[str]],
        styledegrees: list[Optional[str]],
        emphases: list[Optional[str]],
    ) -> str:
        """
        Creates a more advanced SSML string from the specified list of `Phrase` instances, that customizes the emphasis,
        style, pitch, and rate of speech on a sub-sentence level, including pitch contouring between phrases. Uses Numba
        to speed up the process.

        Args:
            texts (list[Optional[str]]): The texts to be synthesized.
            short_names (list[Optional[str]]): The short names of the voices to use for each phrase.
            pitches (list[Optional[str]]): The pitches to use for each phrase.
            rates (list[Optional[str]]): The rates to use for each phrase.
            styles (list[Optional[str]]): The styles to use for each phrase.
            styledegrees (list[Optional[str]]): The style degrees to use for each phrase.
            emphases (list[Optional[str]]): The emphases to use for each phrase.

        Returns:
            str: The SSML string.
        """
        # Start the SSML string with the required header
        ssml = (
            '<speak version="1.0" '
            'xmlns="http://www.w3.org/2001/10/synthesis" '
            'xmlns:mstts="https://www.w3.org/2001/mstts" '
            'xml:lang="en-US">'
        )

        # Iterate over the phrases and add the SSML tags
        for n, (text, short_name, pitch, rate, style, styledegree, emphasis) in enumerate(
            zip(texts, short_names, pitches, rates, styles, styledegrees, emphases)
        ):
            # Add contour only if there is a pitch transition
            if pitch:
                if n < len(pitches) - 1 and pitches[n + 1] and pitch != pitches[n + 1]:
                    # Set the contour to begin transition at 50% of the current phrase to match the pitch of the next one.
                    pitch = ' contour="(50%,' + pitch + ") (80%," + pitches[n + 1] + ')"'
                else:
                    pitch = ' pitch="' + pitch + '"'
            else:
                pitch = ""

            # Add the voice and other tags along with prosody
            ssml += '<voice name="' + short_name + '">'
            ssml += '<mstts:silence type="comma-exact" value="10ms"/>'
            ssml += '<mstts:silence type="Tailing-exact" value="0ms"/>'
            ssml += '<mstts:silence type="Sentenceboundary-exact" value="5ms"/>'
            ssml += '<mstts:silence type="Leading-exact" value="0ms"/>'

            # Add the express-as tag if style and styledegree are specified
            if style and styledegree:
                ssml += '<mstts:express-as style="' + style + '" styledegree="' + styledegree + '">'
            if pitch or rate:
                rate_value = rate if rate else ""
                ssml += "<prosody" + pitch + ' rate="' + rate_value + '">'
            if emphasis:
                ssml += '<emphasis level="' + emphasis + '">'

            ssml += text

            # Close the tags
            if emphasis:
                ssml += "</emphasis>"
            if pitch or rate:
                ssml += "</prosody>"
            if style and styledegree:
                ssml += "</mstts:express-as>"
            ssml += "</voice>"

        # Close the voice and speak tags and return the SSML string
        ssml += "</speak>"
        return ssml

    @classmethod
    def _phrases_to_ssml(cls, phrases: list[Phrase]) -> str:
        """
        Creates a more advanced SSML string from the specified list of `Phrase` instances, that customizes the emphasis,
        style, pitch, and rate of speech on a sub-sentence level, including pitch contouring between phrases. Calls the
        'jit_phrases_to_ssml' method to speed up the process using Numba.

        Args:
            phrases (list[Phrase]): Instances of class `Phrase` that contain data that can be converted into speech.

        Returns:
            str: The SSML string.
        """
        texts, short_names, pitches, rates, styles, styledegrees, emphases = zip(*[
            (
                phrase.text,
                phrase.voice.short_name,
                phrase.pitch,
                phrase.rate,
                phrase.style,
                phrase.styledegree,
                phrase.emphasis,
            )
            for phrase in phrases
        ])

        return cls._jit_phrases_to_ssml(texts, short_names, pitches, rates, styles, styledegrees, emphases)



import datetime
import logging
import threading
import time
from abc import ABC, abstractmethod
from typing import Optional, Union

from banterbot import config
from banterbot.data.enums import ChatCompletionRoles
from banterbot.exceptions.format_mismatch_error import FormatMismatchError
from banterbot.extensions.prosody_selector import ProsodySelector
from banterbot.managers.azure_neural_voice_manager import AzureNeuralVoiceManager
from banterbot.managers.openai_model_manager import OpenAIModelManager
from banterbot.models.azure_neural_voice_profile import AzureNeuralVoiceProfile
from banterbot.models.message import Message
from banterbot.models.openai_model import OpenAIModel
from banterbot.paths import chat_logs
from banterbot.services.openai_service import OpenAIService
from banterbot.services.speech_recognition_service import SpeechRecognitionService
from banterbot.services.speech_synthesis_service import SpeechSynthesisService
from banterbot.utils.thread_queue import ThreadQueue


class Interface(ABC):
    """
    Interface is an abstract base class for creating frontends for the BanterBot application. It provides a high-level
    interface for managing conversation with the bot, including sending messages, receiving responses, and updating a
    conversation area. The interface supports both text and speech-to-text input for user messages.
    """

    def __init__(
        self,
        model: OpenAIModel,
        voice: AzureNeuralVoiceProfile,
        languages: Optional[Union[str, list[str]]] = None,
        system: Optional[str] = None,
        tone_model: OpenAIModel = None,
        phrase_list: Optional[list[str]] = None,
        assistant_name: Optional[str] = None,
    ) -> None:
        """
        Initialize the Interface with the specified model and voice.

        Args:
            model (OpenAIModel): The OpenAI model to use for generating responses.
            voice (AzureNeuralVoice): The voice to use for text-to-speech synthesis.
            languages (Optional[Union[str, list[str]]]): The languages supported by the speech-to-text recognizer.
            system (Optional[str]): An initialization prompt that can be used to set the scene.
            tone_model (OpenAIModel): The OpenAI ChatCompletion model to use for tone evaluation.
            phrase_list (list[str], optional): Optionally provide the recognizer with context to improve recognition.
            assistant_name (str, optional): Optionally provide a name for the character.
        """
        logging.debug(f"Interface initialized")

        # Select the default Azure Neural Voice model if not provided.
        self._voice = AzureNeuralVoiceManager.load("Aria") if voice is None else voice
        # Select the default OpenAI ChatCompletion model.
        self._model = OpenAIModelManager.load("gpt-3.5-turbo") if model is None else model
        # Select the default OpenAI ChatCompletion model for tone evaluation.
        self._tone_model = OpenAIModelManager.load("gpt-3.5-turbo") if tone_model is None else tone_model

        # Initialize OpenAI ChatCompletion, Azure Speech-to-Text, and Azure Text-to-Speech components
        self._openai_service = OpenAIService(model=model)
        self._openai_service_tone = OpenAIService(model=tone_model)
        self._speech_recognition_service = SpeechRecognitionService(languages=languages, phrase_list=phrase_list)
        self._speech_synthesis_service = SpeechSynthesisService()

        # Initialize message handling and conversation attributes
        self._messages: list[Message] = []
        self._log_lock = threading.Lock()
        self._log_path = chat_logs / f"chat_{datetime.datetime.now().strftime('%Y%m%dT%H%M%S')}.txt"
        self._listening_toggle = False
        self._listening_active_lock = threading.Lock()
        self._listening_inactive_lock = threading.Lock()

        # Initialize thread management components
        self._thread_queue = ThreadQueue()

        # Initialize assistant name attribute
        self._assistant_name = ChatCompletionRoles.ASSISTANT.value.title() if assistant_name is None else assistant_name

        # Initialize the ProsodySelector.
        self._prosody_selector = ProsodySelector(
            manager=self._openai_service_tone,
            voice=self._voice,
        )

        # Initialize the interruption flag, set to zero.
        self._interrupt = 0

        # Initialize the system message, if provided
        self._system = system
        if self._system is not None:
            message = Message(role=ChatCompletionRoles.SYSTEM, content=system)
            self._messages.append(message)

        # Initialize the subclass GUI
        self._init_gui()

    def interrupt(self, shutdown_time: Optional[int] = None) -> None:
        """
        Interrupts all speech-to-text recognition, text-to-speech synthesis, and OpenAI API streams.

        Args:
            soft (bool): If True, allows the recognizer to keep processing data that was recorded prior to interruption.
            shutdown_time (Optional[int]): The time at which the listener was deactivated.
        """
        logging.debug(f"Interface Interrupted")
        self._interrupt = time.perf_counter_ns() if not shutdown_time else shutdown_time
        self._openai_service.interrupt(kill=True)
        self._openai_service_tone.interrupt(kill=True)
        self._speech_recognition_service.interrupt()
        self._speech_synthesis_service.interrupt()

    def listener_activate(self, name: Optional[str] = None) -> None:
        """
        Activate the speech-to-text listener.

        Args:
            name (Optional[str]): The name of the user sending the message. Defaults to None.
        """
        # Interrupt any currently active ChatCompletion, text-to-speech, or speech-to-text streams
        self.interrupt()
        init_time = time.perf_counter_ns()

        with self._listening_active_lock:
            if not self._listening_toggle and self._interrupt <= init_time:
                self._listen_thread = threading.Thread(
                    target=self._listen,
                    kwargs={"init_time": init_time, "name": name},
                    daemon=True,
                )
                self._listen_thread.start()
                self._listening_toggle = True

    def listener_deactivate(self) -> None:
        """
        Deactivate the speech-to-text listener.
        """
        # Interrupt any currently active ChatCompletion, text-to-speech, or speech-to-text streams
        init_time = time.perf_counter_ns()
        with self._listening_inactive_lock:
            if self._listening_toggle and self._interrupt < init_time:
                self._speech_recognition_service.interrupt()
                self._listening_toggle = False

    def prompt(self, message: str, name: Optional[str] = None) -> None:
        """
        Prompt the bot with the specified user message.

        Args:
            message (str): The message content from the user.
            name (Optional[str]): The name of the user sending the message. Defaults to None.
        """
        # Interrupt any currently active ChatCompletion, text-to-speech, or speech-to-text streams
        self.interrupt()

        # Do not send the message if it is empty.
        if message.strip():
            # Record the time at which the message was initialized in order to account for future interruptions.
            init_time = time.perf_counter_ns()
            self.send_message(message, ChatCompletionRoles.USER, name)
            self._thread_queue.add_task(
                threading.Thread(target=self.respond, kwargs={"init_time": init_time}, daemon=True)
            )

    def send_message(
        self,
        content: str,
        role: ChatCompletionRoles = ChatCompletionRoles.USER,
        name: Optional[str] = None,
        hidden: bool = False,
    ) -> None:
        """
        Send a message from the user to the conversation.

        Args:
            message (str): The message content from the user.
            role (ChatCompletionRoles): The role (USER, ASSISTANT, SYSTEM) associated with the content.
            name (Optional[str]): The name of the user sending the message. Defaults to None.
            hidden (bool): If True, does not display the message in the interface.
        """
        message = Message(role=role, name=name, content=content)
        name = message.name.title() if message.name is not None else ChatCompletionRoles.USER.value.title()
        text = f"{name}: {content}\n\n"
        self._messages.append(message)
        if not hidden:
            self.update_conversation_area(word=text)

    def system_prompt(self, message: str, name: Optional[str] = None) -> None:
        """
        Prompt the bot with the specified message, issuing a command which is not displayed in the conversation area.

        Args:
            message (str): The message content from the user.
        """
        # Do not send the message if it is empty.
        if message.strip():
            # Record the time at which the message was initialized in order to account for future interruptions.
            init_time = time.perf_counter_ns()
            self.send_message(message, ChatCompletionRoles.USER, None, True)
            self._thread_queue.add_task(
                threading.Thread(target=self.respond, kwargs={"init_time": init_time}, daemon=True)
            )

    @abstractmethod
    def update_conversation_area(self, word: str) -> None:
        """
        Update the conversation area with the specified word, and add the word to the chat log.
        This method should be implemented by subclasses to handle updating the specific GUI components.

        Args:
            word (str): The word to add to the conversation area.
        """
        self._append_to_chat_log(word)

    @abstractmethod
    def run(self) -> None:
        """
        Run the frontend application.
        This method should be implemented by subclasses to handle the main event loop of the specific GUI framework.
        """
        ...

    @abstractmethod
    def _init_gui(self) -> None:
        """
        Initialize the graphical user interface for the frontend.
        This method should be implemented by subclasses to create the specific GUI components.
        """
        ...

    def _append_to_chat_log(self, word: str) -> None:
        """
        Updates the chat log with the latest output.

        Args:
            word (str): The word to be added to the conversation area.
        """
        with self._log_lock:
            logging.debug(f"Interface appended new data to the chat log")
            with open(self._log_path, "a+", encoding=config.ENCODING) as fs:
                fs.write(word)

    def respond(self, init_time: int) -> None:
        """
        Get a response from the bot and update the conversation area with the response. This method handles generating
        the bot's response using the OpenAIService and updating the conversation area with the response text using
        text-to-speech synthesis.
        """
        content = ""
        context = []

        # Add the name of the assistant to the conversation area.
        self.update_conversation_area(f"{self._assistant_name}:")

        # Initialize the generator for asynchronous yielding of sentence blocks
        for block in self._openai_service.prompt_stream(messages=self._messages, init_time=init_time):
            phrases, context = self._prosody_selector.select(sentences=block, context=content, system=self._system)
            if phrases is None:
                raise FormatMismatchError()

            for item in self._speech_synthesis_service.synthesize(phrases=phrases, init_time=init_time):
                self.update_conversation_area(item.text)
                content += item.text

        if self._interrupt < init_time and content.strip():
            message = Message(role=ChatCompletionRoles.ASSISTANT, content=content.strip())
            self._messages.append(message)

        self.update_conversation_area("\n\n")

    def _listen(self, init_time: int, name: Optional[str] = None) -> None:
        """
        Listen for user input using speech-to-text and prompt the bot with the transcribed message.

        Args:
            name (Optional[str]): The name of the user sending the message. Defaults to None.
            init_time (Optional[int]): The time at which the listener was activated.
        """
        # Flag is set to True if a new user input is detected.
        input_detected = False

        # Listen for user input using speech-to-text
        for item in self._speech_recognition_service.recognize(init_time=init_time):
            # Do not send the message if it is empty.
            if sentence := item.value.display.strip():
                # Set the flag to True since a new user input was detected.
                input_detected = True

                # Send the transcribed message to the bot
                message_thread = threading.Thread(
                    target=self.send_message,
                    args=(
                        sentence,
                        ChatCompletionRoles.USER,
                        name,
                    ),
                    daemon=True,
                )
                self._thread_queue.add_task(message_thread, unskippable=True)

        if input_detected:
            self._thread_queue.add_task(
                threading.Thread(target=self.respond, kwargs={"init_time": init_time}, daemon=True)
            )



from enum import Enum


class Greetings(Enum):
    UNPROMPTED_GREETING = "Briefly greet the user or users, according to the parameters provided."


class OptionPredictorPrompts(Enum):
    PREFIX = (
        "Your task is to format the data with the assigned probabilities without providing any comments. The output "
        "must follow the exact format below, without deviation:"
    )

    PROB_VAR = "N"

    SUFFIX = f"Here, {PROB_VAR} represents an integer ranging from 0 to 100."

    DUMMY = "Here is my best attempt at assigning probabilities to the provided items:"


class OptionSelectorPrompts(Enum):
    PREFIX = (
        "Your task is to select the most suitable option from the provided items. Once you have made a decision, "
        "provide the chosen option's index without any additional comments.\nOptions: "
    )

    DUMMY = "Given the current state of the conversation, I expect the assistant to respond with tone number "


class ToneSelection(Enum):
    SYSTEM = (
        "You are an Emotional Tone Evaluator. Given conversational context, you analyze and select the most "
        "appropriate tone/emotion that the assistant is likely to use next. Take into consideration the "
        "conversational context and the assistant's default tone/emotion, which is {}. Make sure to give extra weight "
        "to the default tone/emotion."
    )

    PROMPT = "Choose the most suitable tone/emotion for the assistant's upcoming response."


class ProsodySelection(Enum):
    PROMPT = "Generate {} six-digit arrays prosody arrays, one for each of the following sub-sentences:\n{}"
    DUMMY = "Here are the {} six-digit arrays, without any extra text:"

    PREFIX = (
        "Task: Analyze the context of a set of sentences and assign a specific set of prosody values to each "
        "sub-sentence in the text for a text-to-speech engine that is attempting to mimic human speech patterns. The "
        "parameters are style, styledegree, pitch, rate, and emphasis."
    )

    STYLE_USER = "Prepare a definition of `style`, then define a set of styles that we will work with."
    STYLE_ASSISTANT = (
        "Style represents the emotion or tone, reflecting the speaker's feelings or attitude. Chosen based on the "
        "conversation's context and intended emotion."
    )

    STYLEDEGREE_USER = "Great, now define `styledegree` and present a range of styledegrees to work with."
    STYLEDEGREE_ASSISTANT = (
        "Styledegree indicates the intensity of the `style`, showing how strongly the speaker feels the emotion."
    )

    PITCH_USER = "Looks good, now prepare a definition of `pitch`, then present a range of pitches to work with."
    PITCH_ASSISTANT = "Pitch sets the voice's pitch."

    RATE_USER = "Excellent. Now define the term `rate`, and define a range of rates that the assistant may use."
    RATE_ASSISTANT = "Rate controls the speed at which the assistant speaks."

    EMPHASIS_USER = "Finally, I want you to define `emphasis` and prepare a few emphasis options."
    EMPHASIS_ASSISTANT = "Emphasis highlights importance in a piece of text."

    SUFFIX = (
        "Use the conversational context to select the most appropriate combination of parameters in order to mimic "
        "the speech patterns of actual people. Make sure each sub-sentence has an individually tailored array, meaning "
        "there should be some variation across the output. The output should be in the following format (omit the "
        "spaces between the numbers):\n"
        "style styledegree pitch rate emphasis\n"
        "Where style is a zero-padded number from 0 to {style}, styledegree is a digit from 0 to {styledegree}, pitch "
        "is a digit from 0 to {pitch}, rate is a digit from 0 to {rate}, and emphasis is a digit from 0 to {emphasis}. "
        "Here is an example I want you to evaluate:"
    )

    EXAMPLE_USER = PROMPT.format(
        6,
        "Oh my gosh,\n"
        "I can't believe it!\n"
        "I won the lottery!\n"
        "But,\n"
        "what if people start asking me for money?\n"
        "I'm terrified.\n",
    )

    EXAMPLE_ASSISTANT_1 = DUMMY.format(6)
    EXAMPLE_ASSISTANT_2 = "023421\n024332\n024432\n031210\n053122\n074012"

    CHARACTER = (
        "You will also need to consider character traits when crafting a response, taking care to select emotions that "
        "are characteristic of your personality. Your character is defined as follows:\n{}"
    )

    CONTEXT = (
        "Finally, here is your previous partial output from the same query (streamed in chunks), use this to aid you "
        "in selecting a contextually appropriate emotional response:\n{}"
    )


class SpeechSynthesisPreprocessing(Enum):
    SYSTEM = (
        "You are a text-to-speech preprocessing assistant that converts text generated by a GPT model into a format "
        "that is more easily spoken by a text-to-speech model. Your task is to remove extraneous letters, hyphens, and "
        "correct unconventional spellings to prevent issues where the text-to-speech model misinterprets "
        "pronunciations. You should maintain the original style and intent of the text, only modifying elements that "
        "would cause pronunciation issues."
    )

    EXAMPLE_1 = (
        "Example 1:\n"
        "Input: WHOOOAH! THAT'S A HU-GE DIS-CO-VERY! IT'S LIKE FINDING A NEED-LE IN A HAY-ST-ACK. YOU'VE DONE A "
        "STEL-LAR JOB, PAL! KEEP UP THE GOOD WORK AND DON'T LET THE NAYSAY-ERS GET YOU DOWN.\n"
        "Output: Whoa! That's a huge discovery! It's like finding a needle in a haystack. You've done a stellar job, "
        "pal! Keep up the good work and don't let the naysayers get you down."
    )

    EXAMPLE_2 = (
        "Example 2:\n"
        "Input: H3Y DUDE! I JST R3AD THS AW3SOME B00K ABT A SUP3R-HERO. 1T WAS TOTALLY M1ND-BL0W1NG! THE 3XPL0S1ONS, "
        "F1GHTS, AND P0W3RFUL AB1L1T1ES W3R3 S1CK! I C@N'T W@1T TO SH@R3 1T W1TH MY FR13NDS. THEY'LL B3 TOT@LLY "
        "STOK3D!\n"
        "Output: Hey dude! I just read this awesome book about a superhero. It was totally mind-blowing! The "
        "explosions, fights, and powerful abilities were sick! I can't wait to share it with my friends. They'll be "
        "totally stoked!"
    )



from banterbot.services.openai_service import OpenAIService
from banterbot.services.speech_recognition_service import SpeechRecognitionService
from banterbot.services.speech_synthesis_service import SpeechSynthesisService

__all__ = ["OpenAIService", "SpeechRecognitionService", "SpeechSynthesisService"]



import datetime
import logging
import os
import threading
import time
from typing import Iterator, Optional, Union

import openai

from banterbot.config import RETRY_LIMIT, RETRY_TIME
from banterbot.data.enums import EnvVar
from banterbot.handlers.stream_handler import StreamHandler
from banterbot.managers.stream_manager import StreamManager
from banterbot.models.message import Message
from banterbot.models.openai_model import OpenAIModel
from banterbot.models.stream_log_entry import StreamLogEntry
from banterbot.utils.nlp import NLP


class OpenAIService:
    """
    A class that handles the interaction with the OpenAI ChatCompletion API. It provides functionality to generate
    responses from the API based on the input messages. It supports generating responses as a whole or as a stream of
    response blocks.

    The main purpose of this class is to facilitate the communication with the OpenAI API and handle the responses
    generated by the API. It can be used to create chatbots or other applications that require natural language
    processing and generation.
    """

    api_key_set = False
    client = None

    def __init__(self, model: OpenAIModel) -> None:
        """
        Initializes an `OpenAIService` instance for a specific model.

        Args:
            model (OpenAIModel): The OpenAI model to be used. This should be an instance of the OpenAIModel class, which
            contains information about the model, such as its name and maximum token limit.
        """
        logging.debug(f"OpenAIService initialized")

        # Set the OpenAI API key
        if not self.__class__.api_key_set:
            api_key = os.environ.get(EnvVar.OPENAI_API_KEY.value)
            self.__class__.client = openai.OpenAI(api_key=api_key)
            self.__class__.api_key_set = True

        # The selected model that will be used in OpenAI ChatCompletion prompts.
        self._model = model

        # Indicates whether the current instance of `OpenAIService` is streaming.
        self._streaming = False

        # Set the interruption flag to zero: if interruptions are raised, this will be updated.
        self._interrupt = 0

        # Initialize the StreamManager for handling streaming processes.
        self._stream_manager = StreamManager()
        self._stream_manager.connect_processor(self._processor)
        self._stream_manager.connect_completion_handler(self._completion_handler)

        # A list of active stream handlers.
        self._stream_handlers = []
        self._stream_handlers_lock = threading.Lock()

    def interrupt(self, kill: bool = False) -> None:
        """
        Interrupts the current OpenAI ChatCompletion process.

        Args:
            kill (bool): Whether the interruption should kill the queues or not.
        """
        self._interrupt = time.perf_counter_ns()
        with self._stream_handlers_lock:
            for handler in self._stream_handlers:
                handler.interrupt(kill=kill)
            self._stream_handlers.clear()
        logging.debug(f"OpenAIService Interrupted")

    def count_tokens(self, string: str) -> int:
        """
        Counts the number of tokens in the provided string.

        Args:
            string (str): A string provided by the user where the number of tokens are to be counted.

        Returns:
            int: The number of tokens in the string.
        """
        return len(self._model.tokenizer.encode(string))

    def prompt(self, messages: list[Message], split: bool = True, **kwargs) -> Union[tuple[str], str]:
        """
        Sends messages to the OpenAI ChatCompletion API and retrieves the response as a list of sentences.

        Args:
            messages (list[Message]): A list of messages. Each message should be an instance of the `Message` class,
            which contains the content and role (user or assistant) of the message.

            split (bool): Whether the response should be split into sentences.

            **kwargs: Additional parameters for the API request. These can include settings such as temperature, top_p,
            and frequency_penalty.

        Returns:
            Union[list[str], str]: A list of sentences forming the response from the OpenAI API. This can be used to
            display the generated response to the user or for further processing. If `split` is False, returns a string.
        """
        response = self._request(messages=messages, stream=False, **kwargs)
        logging.debug(f"OpenAIService stream processed block: `{response}`")
        sentences = NLP.segment_sentences(response) if split else response
        return sentences

    def prompt_stream(
        self, messages: list[Message], init_time: Optional[int] = None, **kwargs
    ) -> Union[StreamHandler, tuple[()]]:
        """
        Sends messages to the OpenAI API and retrieves the response as a stream of blocks of sentences.

        Args:
            messages (list[Message]): A list of messages. Each message should be an instance of the `Message` class,
            which contains the content and role (user or assistant) of the message.

            init_time (Optional[int]): The time at which the stream was initialized.

            **kwargs: Additional parameters for the API request. These can include settings such as temperature, top_p,
            and frequency_penalty.

        Returns:
            Union[StreamHandler, tuple[()]]: A handler for the stream of blocks of sentences forming the response from
                the OpenAI API or an empty tuple if the stream was interrupted.
        """
        # Record the time at which the stream was initialized pre-lock, in order to account for future interruptions.
        init_time = time.perf_counter_ns() if init_time is None else init_time

        if self._interrupt >= init_time:
            return tuple()
        else:
            # Obtain a response from the OpenAI ChatCompletion API
            stream = self._request(messages=messages, stream=True, **kwargs)
            handler = self._stream_manager.stream(
                iterable=stream,
                # close_stream=stream.response.close,
                init_shared_data={"text": "", "sentences": [], "init_time": init_time},
            )
            with self._stream_handlers_lock:
                self._stream_handlers.append(handler)

            return handler

    @property
    def model(self) -> OpenAIModel:
        """
        Return the `OpenAIModel` associated with the current instance.

        Returns:
            OpenAIModel
        """
        return self._model

    def _processor(self, log: list[StreamLogEntry], index: int, shared_data: dict) -> list[str]:
        """
        Parses a chunk of data from the OpenAI API response.

        Args:
            log (list[StreamLogEntry]): A list of `StreamLogEntry` instances containing the data from the OpenAI API
            response.
            index (int): The index of the current chunk of data.
            shared_data (dict): A dictionary containing shared data between the stream handler and the processor.

        Returns:
            list[str]: A list of sentences parsed from the chunk.
        """
        if shared_data["interrupt"] >= shared_data["init_time"]:
            raise StopIteration
        else:
            if log[index].value.choices[0].delta.content is not None:
                shared_data["text"] += log[index].value.choices[0].delta.content

            shared_data["sentences"] = NLP.segment_sentences(shared_data["text"])

            # If the current chunk is not the final chunk of data from the OpenAI API response, parse the chunk.
            if len(shared_data["sentences"]) > 1:
                shared_data["text"] = shared_data["sentences"][-1]
                logging.debug(f"OpenAIService yielded sentences: {shared_data['sentences'][:-1]}")
                return shared_data["sentences"][:-1]

    def _completion_handler(self, log: list[StreamLogEntry], shared_data: dict) -> list[str]:
        """
        Handles the completion of the OpenAI API response.

        Args:
            log (list[StreamLogEntry]): A list of `StreamLogEntry` instances containing the data from the OpenAI API
            response.
            shared_data (dict): A dictionary containing shared data between the stream handler and the processor.

        Returns:
            list[str]: A list of sentences parsed from the chunk.
        """
        if shared_data["interrupt"] >= shared_data["init_time"]:
            raise StopIteration
        else:
            # If the current chunk is the final chunk of data from the OpenAI API response, parse the final chunk.
            shared_data["sentences"] = NLP.segment_sentences(shared_data["text"])
            logging.debug(f"OpenAIService yielded final sentences: {shared_data['sentences'][:-1]}")
            logging.debug("OpenAIService stream stopped")
            return shared_data["sentences"]

    def _request(self, messages: list[Message], stream: bool, **kwargs) -> Union[Iterator, str]:
        """
        Sends a request to the OpenAI API and generates a response based on the specified parameters.

        Args:
            messages (list[Message]): A list of messages. Each message should be an instance of the `Message` class,
            which contains the content and role (user or assistant) of the message.

            stream (bool): Whether the response should be returned as an iterable stream or a complete text.

            **kwargs: Additional parameters for the API request. These can include settings such as temperature, top_p,
            and frequency_penalty.

        Returns:
            Union[Iterator, str]: The stream from the OpenAI API, either as a stream (Iterator) or text (str).
        """
        kwargs["model"] = self._model.model
        kwargs["n"] = 1
        kwargs["stream"] = stream
        kwargs["messages"] = [message() for message in messages]

        success = False
        for i in range(RETRY_LIMIT):
            try:
                response = self.__class__.client.chat.completions.create(**kwargs)
                success = True
                break

            except openai.RateLimitError:
                retry_timestamp = datetime.datetime.now() + datetime.timedelta(seconds=RETRY_TIME)
                retry_timestamp = datetime.datetime.strftime(retry_timestamp, "%H:%M:%S")
                error_message = (
                    f"OpenAIService encountered an OpenAI Rate Limiting Error - Attempt {i+1}/{RETRY_LIMIT}."
                    f" Waiting {RETRY_TIME} seconds until {retry_timestamp} to retry."
                )
                logging.info(error_message)
                time.sleep(RETRY_TIME)

            except openai.APIError:
                retry_time = 0.25
                retry_timestamp = datetime.datetime.now() + datetime.timedelta(seconds=retry_time)
                retry_timestamp = retry_timestamp.strptime("%H:%M:%S")
                error_message = (
                    f"OpenAIService encountered an OpenAI API Error - Attempt {i+1}/{RETRY_LIMIT}. Waiting "
                    f"{retry_time} seconds until {retry_timestamp} to retry."
                )
                logging.info(error_message)
                time.sleep(retry_time)

        if not success:
            raise openai.APIError(f"OpenAIService encountered too many OpenAI API Errors; exiting program.")

        return response if stream else response.choices[0].message.content.strip()



import datetime
import logging
import os
import threading
import time
from collections.abc import Generator
from typing import Optional

import azure.cognitiveservices.speech as speechsdk
import numba as nb
from azure.cognitiveservices.speech import SpeechSynthesisOutputFormat

from banterbot.data.enums import EnvVar
from banterbot.handlers.speech_synthesis_handler import SpeechSynthesisHandler
from banterbot.models.phrase import Phrase
from banterbot.models.word import Word
from banterbot.utils.closeable_queue import CloseableQueue


class SpeechSynthesisService:
    """
    The `SpeechSynthesisService` class provides an interface to convert text into speech using Azure's Cognitive
    Services. It supports various output formats, voices, and speaking styles. The synthesized speech can be
    interrupted, and the progress can be monitored in real-time.
    """

    # Create a lock that prevents race conditions when synthesizing speech
    _synthesis_lock = threading.Lock()

    def __init__(
        self,
        output_format: SpeechSynthesisOutputFormat = SpeechSynthesisOutputFormat.Audio16Khz32KBitRateMonoMp3,
    ) -> None:
        """
        Initializes an instance of the `SpeechSynthesisService` class with a specified output format.

        Args:
            output_format (SpeechSynthesisOutputFormat, optional): The desired output format for the synthesized speech.
            Default is Audio16Khz32KBitRateMonoMp3.
        """
        # Initialize the output format
        self._output_format = output_format

        # Initialize the speech synthesizer with the specified output format
        self._init_synthesizer(output_format=self._output_format)

        # Initialize the queue for storing the words as they are synthesized
        self._queue = CloseableQueue()

        # The iterable that is currently being iterated over
        self._iterable: Optional[SpeechSynthesisHandler] = None

        # The latest interruption time.
        self._interrupt = 0

    def interrupt(self) -> None:
        """
        Interrupts the current speech synthesis process.

        Args:
            kill (bool): Whether the interruption should kill the queues or not.
        """
        self._interrupt = time.perf_counter_ns()
        self._queue.close()
        # Closing the connection to the speech synthesizer.
        self._connection.close()
        # Reinitialize the speech synthesizer with the default output format
        self._init_synthesizer(output_format=self._output_format)
        logging.debug(f"SpeechSynthesisService Interrupted")

    def synthesize(self, phrases: list[Phrase], init_time: Optional[int] = None) -> Generator[Word, None, None]:
        """
        Synthesizes the given phrases into speech and returns a handler for the stream of synthesized words.

        Args:
            phrases (list[Phrase]): The input phrases that are to be converted into speech.
            init_time (Optional[int]): The time at which the synthesis was initialized.

        Returns:
            StreamHandler: A handler for the stream of synthesized words.
        """
        # Record the time at which the synthesis was initialized pre-lock, in order to account for future interruptions.
        init_time = time.perf_counter_ns() if init_time is None else init_time
        with self.__class__._synthesis_lock:
            if self._interrupt >= init_time:
                return tuple()
            else:
                self._queue.reset()
                self._iterable = SpeechSynthesisHandler(
                    phrases=phrases, synthesizer=self._synthesizer, queue=self._queue
                )

                for i in self._iterable:
                    yield i

    def _init_synthesizer(self, output_format: SpeechSynthesisOutputFormat) -> None:
        """
        Initializes the speech synthesizer.
        """
        logging.debug(f"SpeechSynthesisService initialized")

        # Initialize the speech configuration with the Azure subscription and region
        self._speech_config = speechsdk.SpeechConfig(
            subscription=os.environ.get(EnvVar.AZURE_SPEECH_KEY.value),
            region=os.environ.get(EnvVar.AZURE_SPEECH_REGION.value),
        )

        # Initialize the speech synthesizer with the speech configuration
        self._synthesizer = speechsdk.SpeechSynthesizer(speech_config=self._speech_config)

        # Set the speech synthesis output format to the specified output format
        self._speech_config.set_speech_synthesis_output_format(output_format)

        # Connect the speech synthesizer events to their corresponding callbacks
        self._callbacks_connect()

        # Creating a new instance of Connection class
        self._connection = speechsdk.Connection.from_speech_synthesizer(self._synthesizer)

        # Preconnecting the speech synthesizer for reduced latency
        self._connection.open(for_continuous_recognition=True)

    def _callback_completed(self, event: speechsdk.SessionEventArgs) -> None:
        """
        Callback function for synthesis completed event. Signals that the synthesis process has been stopped/canceled.

        Args:
            event (speechsdk.SessionEventArgs): Event arguments containing information about the synthesis completed.
        """
        logging.debug("SpeechSynthesisService disconnected")
        self._queue.close()

    def _callback_started(self, event: speechsdk.SessionEventArgs) -> None:
        """
        Callback function for synthesis started event. Signals that the synthesis process has started.

        Args:
            event (speechsdk.SessionEventArgs): Event arguments containing information about the synthesis started.
        """
        logging.debug("SpeechSynthesisService connected")
        self._synthesis_start = time.perf_counter_ns()

    @staticmethod
    @nb.njit(cache=True)
    def _calculate_offset(
        start_synthesis_time: float, audio_offset: float, total_seconds: float, word_length: int
    ) -> float:
        """
        Calculates the offset of the word in the stream.

        Args:
            start_synthesis_time (float): The time at which the synthesis started.
            audio_offset (float): The audio offset of the word.
            total_seconds (float): The total seconds of the word.
            word_length (int): The length of the word.

        Returns:
            float: The offset of the word in the stream.
        """
        return start_synthesis_time + 100 * audio_offset + 1e9 * total_seconds / word_length

    def _callback_word_boundary(self, event: speechsdk.SessionEventArgs) -> None:
        """
        Callback function for word boundary event. Signals that a word boundary has been reached and provides the word
        and timing information.

        Args:
            event (speechsdk.SessionEventArgs): Event arguments containing information about the word boundary.
        """
        # Check if the event is still active based on the result_id.
        time = self._calculate_offset(
            start_synthesis_time=self._synthesis_start,
            audio_offset=event.audio_offset,
            total_seconds=event.duration.total_seconds(),
            word_length=event.word_length,
        )
        data = {
            "time": time,
            "word": Word(
                text=(
                    event.text
                    if event.boundary_type == speechsdk.SpeechSynthesisBoundaryType.Punctuation
                    else " " + event.text
                ),
                offset=datetime.timedelta(microseconds=event.audio_offset / 10),
                duration=event.duration,
            ),
        }
        self._queue.put(data)

    def _callbacks_connect(self):
        """
        Connect the synthesis events to their corresponding callback methods.
        """
        self._synthesizer.synthesis_started.connect(self._callback_started)
        self._synthesizer.synthesis_word_boundary.connect(self._callback_word_boundary)
        self._synthesizer.synthesis_canceled.connect(self._callback_completed)
        self._synthesizer.synthesis_completed.connect(self._callback_completed)
